/*******************************************
   Setup code to collect OV7670 data using
   CSI hardware and DMA.  It can run in the backgroud
   and the camera's RGB565 data appears 'magically' in the
   assigned data buffers.   This initial version does
   full VGA resolution at up to 30FPS.  Since each pixel
   of RGB565 data requires two bytes, a full frame of VGA
   requires 640x480x2 bytes, or 614400 bytes.  The SCI can
   alternate between two buffers. Since a 614KB
   buffer is too large for a single segment of the T4.1 internal
   RAM, you will need to have the 8MB PSRAM installed.

   MJB  10/26/2020
*/

/**************************************************
   These defines aren't used, but they're a handy
   reminder of how the )V7670 gets wired to the T4.1
  #define RST_PIN 14 // hardware reset for my board

  #define SCL_PIN 19
  #define SDA_PIN 18

  #define ECLK_PIN 41// CSI_MCLK AD_B1_05
  #define VD_PIN 17 //CSI_VSYNC AD_B1_06
  #define HD_PIN 16 // CSI_HSYNC AD_B1_07
  #define DCLK_PIN 40 // CSI_PIXCLK AD_B1_04

  #define D0_PIN 27// AD_B1_15
  #define D1_PIN 26// AD_B1_14
  #define D2_PIN 39// AD_B1_13
  #define D3_PIN 38// AD_B1_12
  #define D4_PIN 21// AD_B1_11
  #define D5_PIN 20// AD_B1_10
  #define D6_PIN 23// AD_B1_09
  #define D7_PIN 22// AD_B1_08
**********************************************/
#include "CSI_41.h"
// define the bits in the CSISR status register
#define BIT_SOF_INT 16
#define BIT_EOF_INT 17
uint32_t last_fc = 0;



uint32_t csisr_state = 0;

void CSI_Start(void) {
  CSI_CSICR18 |= (CSI_CSICR18_CSI_ENABLE //CSI_ENABLE;
                  | CSI_CSICR18_BASEADDR_SWITCH_EN // CSI switches base addresses at frame start
                  | CSI_CSICR18_BASEADDR_CHANGE_ERROR_IE); //  Enables base address error interrupt
}

void CSI_Stop(void) {
  CSI_CSICR18 &= ~CSI_CSICR18_CSI_ENABLE; ///  clear the enable bit
}

// IRQ Handler for CSI Interrupt
void CSIService(void) {
  last_csistat =  CSI_CSISR;

  IMARKTOGGLE
  if (last_csistat & CSI_CSISR_SOF_INT) {
    csiIrqCount++;
    //enable DMA requests
    //   CSI_CSICR3 = (CSI_CSICR3 & ~(0b111 << 4)) | (0b010 << 4); // RxFF_LEVEL

  }
  if (last_csistat & CSI_CSISR_DMA_TSF_DONE_FB2) fb2count++; // increment buffer 2 count
  if (last_csistat & CSI_CSISR_DMA_TSF_DONE_FB1) fb1count++; // increment buffer 1 count
  // toggle IMARK with DMA End flags

  if (last_csistat & CSI_CSISR_BASEADDR_CHHANGE_ERROR) {
    bswitcherrors++; // buffer switch before DMA done
  }
  CSI_CSISR = last_csistat; // clear all interrupts
}

/************************************************
  Set CSI internal CCM clock to 24 MHz or 12 MHz
  Max VGA Frame rate at 12MHz is 15FPS.
  At 24MHz, you can get 30 VGA FPS, but the system seems
  to fail to store the first byte of data, resulting in
  images that look like false-color thermal images.
* ************************************************/
void CSI_InitMCLK(uint16_t MHz) {

  CCM_CCGR2 &= (~0b1100);  // turn off csi clock

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_05 = 0b100; // alt4
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_05 = 0x18U;  // 50MHz speed, DSE_3
  // anything but 24MHz defaults to 12MHz for now
  if (MHz  == 24) {
    CCM_CSCDR3 = CCM_CSCDR2_LCDIF_CLK_SEL(0) | CCM_CSCDR3_CSI_PODF(0);  // set csi clock source and divide by 1 = 24 MHz
    masterfreq = 24;
  } else {
    CCM_CSCDR3 = CCM_CSCDR2_LCDIF_CLK_SEL(0) | CCM_CSCDR3_CSI_PODF(1);  // set csi clock source and divide by 2 = 12 MHz
    masterfreq = 12;
  }
  CCM_CCGR2 = (CCM_CCGR2 & (~0b1100)) | 0b1100;  // turn on csi clock

}


// clearing the FIFO is complicated in that it can only be done
// when the FCC bit is 0.  This bit is normally set so that the
// FIFOs automatically transfer to the external memory as required.
void CSI_ClearFIFOs(void) {
  uint32_t creg, maskbits;
  // Clear the FIFOs--which can only be done when FCC is 0
  creg = CSI_CSICR1;
  CSI_CSICR1 = (creg & ~CSI_CSICR1_FCC); // clear FCC bit
  maskbits = CSI_CSICR1_CLR_STATFIFO | CSI_CSICR1_CLR_RXFIFO; // bits for clearing stat and rx FIFOs
  CSI_CSICR1 = creg & ( ~CSI_CSICR1_FCC | maskbits); // keep FCC off and clear FIFOs ??? 
  // wait while clear completes
  while ((CSI_CSICR1 & maskbits) != 0) {}
  CSI_CSICR1 = creg;  // restore original status
}

/*****************************************************************
   Set up the CSI registers.  For now we assume RGB565 mode
   which uses 2 bytes per pixel and needs 61440 bytes per frame.
   That means that the frame buffers have to be in the EXTMEM
   PSRAM.  If you pass the same buffer pointer in fb1 and fb2, the
   CSI will happily overwrite that buffer with each new frame.
********************************************************************/
void CSI_Setup(uint16_t fwidth, uint16_t fheight, uint8_t *fb1, uint8_t *fb2) {
  uint32_t creg;
  NVIC_DISABLE_IRQ(IRQ_CSI);
  // set initial values in CSICR1
  CSI_CSICR1 |= (
                  CSI_CSICR1_SOF_POL |    // 1 for  VSYNC rising edge
                  CSI_CSICR1_FCC  |       // FCC_MASK
                  CSI_CSICR1_HSYNC_POL |  // HSYNC_POL 1 for active high
                  CSI_CSICR1_GCLK_MODE  | // 1= GCLK_MODE PCLK only valid when HSYNC high
                  CSI_CSICR1_REDGE);      // Use pixel clock rising edge

  // Set Image Parameters--width in bytes and height in rows
  // bytes in row goes into upper 16 bits, height in lower 16 bits.
  CSI_CSIIMAG_PARA = CSI_CSIIMAG_PARA_IMAGE_WIDTH(2 * fwidth ) | CSI_CSIIMAG_PARA_IMAGE_HEIGHT(fheight);

  CSI_CSICR3 =  CSI_CSICR3_FRMCNT_RST |     // clear the frame counter
                CSI_CSICR3_DMA_REQ_EN_RFF;  //  Enable RxFIFO DMA request
  CSI_CSICR2 |= CSI_CSICR2_DMA_BURST_TYPE_RFF(0b11);  // RxFIFO DMA_BURST_TYPE increment by 16

  CSI_CSICR18 = (CSI_CSICR18 & ~CSI_CSICR18_MASK_OPTION(0b11)) | CSI_CSICR18_MASK_OPTION(0b10);  // CSI_CSICR18_MASK_OPTION(2)

  // CSI_CSICR2 skip count not needed as frame width matches buffer width
  // But it may be necessary with smaller output formats
  CSI_CSIDMASA_FB1 = (uint32_t)(fb1);
  CSI_CSIDMASA_FB2 = (uint32_t)(fb2);

  // enable interrupt flags.  This needs to be done to read status, even if you
  // don't enable the CSI interrupt in the NVIC.   I enable the interrupt
  // so that I can collect data on errors and the number of frames written
  // in each buffer.
  CSI_CSICR1 |= (
                  CSI_CSICR1_FB1_DMA_DONE_INTEN |   // FB1_DMA_DONE_INTEN
                  CSI_CSICR1_FB2_DMA_DONE_INTEN  |  // FB2_DMA_DONE_INTEN
                  CSI_CSICR1_SOF_INTEN  );          //  SOF Interrupt enable

  CSI_ClearFIFOs();

  // Reset the Receive data DMA hardware
  // Why they call it 'reflashing' is a mystery to me!
  CSI_CSICR3 |= CSI_CSICR3_DMA_REFLASH_RFF;  // CSI_CSICR3_DMA_REFLASH_RFF_MASK;
  while (CSI_CSICR3 & CSI_CSICR3_DMA_REFLASH_RFF);  // wait until reflash done

  // first clear, then set the RxFIFO full level to 16 double words or 64 bytes
  // I think this level is chosen to match the FIFO size of the FlexSPI used for
  // the external PSRAM.
  CSI_CSICR3 = (CSI_CSICR3 & ~CSI_CSICR3_RxFF_LEVEL(0b111)) | CSI_CSICR3_RxFF_LEVEL(0b010); // RxFF_LEVEL 16 double words

  creg = CSI_CSISR;
  CSI_CSISR = creg;  // write back 1's to clear all interrupts

  CSI_Start();
  attachInterruptVector(IRQ_CSI, &CSIService );
  NVIC_ENABLE_IRQ(IRQ_CSI);
}


// Connect the CSI I/O to the external pins of the T4.1
// change the input pins on clocks to have hysteresis  MJB 10/24/2020
void CSI_InitIOMux(void) {
  // VSYNC
  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_06 = 0x4U;
  IOMUXC_CSI_VSYNC_SELECT_INPUT = 0x1U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_06 = 0x00010000U;
  /* Alt VSYNC on pin 34 B1_13 ALT2
    IOMUXC_SW_MUX_CTL_PAD_GPIO_B1_13 = 0x2U;
    IOMUXC_CSI_VSYNC_SELECT_INPUT = 0x2U;
    IOMUXC_SW_PAD_CTL_PAD_GPIO_B1_13 = 0x0U; */
  // HSYNC
  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_07 = 0x4U;
  IOMUXC_CSI_HSYNC_SELECT_INPUT = 0x1U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_07 = 0x00010000U;
  // DCLK or Pixel Clk
  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_04 = 0x4U;
  IOMUXC_CSI_PIXCLK_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_04 = 0x00010000U;
  // D7-D0
  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_08 = 0x4U;
  IOMUXC_CSI_DATA09_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_08 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_09 = 0x4U;
  IOMUXC_CSI_DATA08_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_09 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_10 = 0x4U;
  IOMUXC_CSI_DATA07_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_10 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_11 = 0x4U;
  IOMUXC_CSI_DATA06_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_11 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_12 = 0x4U;
  IOMUXC_CSI_DATA05_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_12 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_13 = 0x4U;
  IOMUXC_CSI_DATA04_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_13 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_14 = 0x4U;
  IOMUXC_CSI_DATA03_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_14 = 0x0U;

  IOMUXC_SW_MUX_CTL_PAD_GPIO_AD_B1_15 = 0x4U;
  IOMUXC_CSI_DATA02_SELECT_INPUT = 0x0U;
  IOMUXC_SW_PAD_CTL_PAD_GPIO_AD_B1_15 = 0x0U;
}


//  A handy function to display the contents of the CSI registers
#define PV(N) Serial.print(#N); Serial.print(": "); Serial.println(N, HEX);
void print_csi_registers() {
  Serial.println("==== CSI ====");
  PV(CSI_CSICR1)
  PV(CSI_CSICR2)
  PV(CSI_CSICR3)
  PV(CSI_CSISTATFIFO)
  PV(CSI_CSIRFIFO)
  PV(CSI_CSIRXCNT)
  PV(CSI_CSISR)
  PV(CSI_CSIDMASA_STATFIFO)
  PV(CSI_CSIDMATS_STATFIFO)
  PV(CSI_CSIDMASA_FB1)
  PV(CSI_CSIDMASA_FB2)
  PV(CSI_CSIFBUF_PARA)
  PV(CSI_CSIIMAG_PARA)
  PV(CSI_CSICR18)
  PV(CSI_CSICR19)
  PV(CSI_CSIRFIFO)
}
