/*******************************************
   Setup code to collect OV7670 data using
   CSI hardware and DMA.  It can run in the backgroud
   and the camera's RGB565 data appears 'magically' in the
   assigned data buffers.   This initial version does
   full VGA resolution at up to 30FPS.  Since each pixel
   of RGB565 data requires two bytes, a full frame of VGA
   requires 640x480x2 bytes, or 614400 bytes.  The SCI can
   alternate between two buffers. Since a 614KB
   buffer is too large for a single segment of the T4.1 internal
   RAM, you will need to have the 8MB PSRAM installed.

   MJB  10/26/2020
*/

/**************************************************
   These defines aren't used, but they're a handy
   reminder of how the )V7670 gets wired to the T4.1
  #define RST_PIN 14 // hardware reset for my board

  #define SCL_PIN 19
  #define SDA_PIN 18

  #define ECLK_PIN 41// CSI_MCLK AD_B1_05
  #define VD_PIN 17 //CSI_VSYNC AD_B1_06
  #define HD_PIN 16 // CSI_HSYNC AD_B1_07
  #define DCLK_PIN 40 // CSI_PIXCLK AD_B1_04

  #define D0_PIN 27// AD_B1_15
  #define D1_PIN 26// AD_B1_14
  #define D2_PIN 39// AD_B1_13
  #define D3_PIN 38// AD_B1_12
  #define D4_PIN 21// AD_B1_11
  #define D5_PIN 20// AD_B1_10
  #define D6_PIN 23// AD_B1_09
  #define D7_PIN 22// AD_B1_08
**********************************************/
#ifndef _CSI_41_H_
#define _CSI_41_H_
#include <Arduino.h>
#include <Wire.h>

// define the bits in the CSISR status register
#define BIT_SOF_INT 16
#define BIT_EOF_INT 17
extern uint32_t last_fc;
extern uint32_t csisr_state;

extern void CSI_Start(void);

extern void CSI_Stop(void);
extern void CSIService(void);
extern void CSI_InitMCLK(uint16_t MHz);
extern void CSI_Setup(uint16_t fwidth, uint16_t fheight, uint8_t *fb1, uint8_t *fb2);
extern void CSI_InitIOMux(void);
extern void print_csi_registers();

enum PixelFormat {
  RGB565,
  BAYERRGB,
  YUV422
};

enum Resolution {
  VGA_640x480,
  QVGA_320x240,
  QQVGA_160x120
};


// Defines in OV7670_MB.cpp

// Read a single uint8_t from address and return it as a uint8_t
extern uint8_t cameraReadRegister(uint8_t reg);

// Writes a single uint8_t (data) into address
extern void cameraWriteRegister(uint8_t reg, uint8_t data);
extern void cameraWriteRegList(struct regval_entry reglist[]);
extern void cameraReset();
extern void cameraReadAll(uint8_t rvals[]);
extern void cameraShowAll(uint8_t rvals[]);
extern void cameraBegin(uint16_t camwidth, uint16_t camheight, uint8_t *fb1, uint8_t *fb2);
extern void cameraClock(int mhz); // Gets XCLK from masterfreq in CSI_41
extern bool cameraReady();
extern bool cameraSetResFormat(uint8_t res, uint8_t format);
extern bool cameraResetSettings();

// Stuff from main ino file
extern uint16_t FRAME_WIDTH;
extern uint16_t FRAME_HEIGHT;
//#define FRAME_WIDTH 640l
//#define FRAME_HEIGHT 480l

#define FB_WIDTH 640l
#define FB_HEIGHT 480l
#define FB_COUNT 2  // two frames in EXTMEM

#define FRAMEBYTES (FRAME_WIDTH * FRAME_HEIGHT *2)
#define FRAMEBUFFBYTES (FB_WIDTH *FB_HEIGHT *2)

extern uint8_t fb1[FRAMEBUFFBYTES] EXTMEM;
extern uint8_t fb2[FRAMEBUFFBYTES] EXTMEM;

extern uint32_t masterfreq;
extern volatile uint32_t last_csistat, csiIrqCount;
extern volatile uint32_t fb1count, fb2count;
extern volatile uint32_t bswitcherrors;
extern uint32_t framecount;

//const int ledpin    = 13;
extern const int imarkpin;
extern const int pinCamReset;


//#define  LEDON digitalWriteFast(ledpin, HIGH); // Also marks IRQ handler timing
//#define  LEDOFF digitalWriteFast(ledpin, LOW);
//#define  LEDTOGGLE digitalToggleFast(ledpin);

#define  IMARKHI digitalWriteFast(imarkpin, HIGH); // Also marks IRQ handler timing
#define  IMARKLO digitalWriteFast(imarkpin, LOW);
#define  IMARKTOGGLE digitalToggleFast(imarkpin);

// these are added to IMXRT - but not in release yet...
#ifndef  CSI_CSICR1_SWAP16_EN
#define CSI_CSICR1_SWAP16_EN                    ((uint32_t)(1<<31))
#define CSI_CSICR1_EXT_VSYNC                    ((uint32_t)(1<<30))
#define CSI_CSICR1_EOF_INT_EN                   ((uint32_t)(1<<29))
#define CSI_CSICR1_PrP_IF_EN                    ((uint32_t)(1<<28))
#define CSI_CSICR1_CCIR_MODE                    ((uint32_t)(1<<27))
#define CSI_CSICR1_COF_INT_EN                   ((uint32_t)(1<<26))
#define CSI_CSICR1_SF_OR_INTEN                  ((uint32_t)(1<<25))
#define CSI_CSICR1_RF_OR_INTEN                  ((uint32_t)(1<<24))
#define CSI_CSICR1_SFF_DMA_DONE_INTEN           ((uint32_t)(1<<22))
#define CSI_CSICR1_STATFF_INTEN                 ((uint32_t)(1<<21))
#define CSI_CSICR1_FB2_DMA_DONE_INTEN           ((uint32_t)(1<<20))
#define CSI_CSICR1_FB1_DMA_DONE_INTEN           ((uint32_t)(1<<19))
#define CSI_CSICR1_RXFF_INTEN                   ((uint32_t)(1<<18))
#define CSI_CSICR1_SOF_POL                      ((uint32_t)(1<<17))
#define CSI_CSICR1_SOF_INTEN                    ((uint32_t)(1<<16))
#define CSI_CSICR1_HSYNC_POL                    ((uint32_t)(1<<11))
#define CSI_CSICR1_CCIR_EN                      ((uint32_t)(1<<10))
#define CSI_CSICR1_FCC                          ((uint32_t)(1<<8))
#define CSI_CSICR1_PACK_DIR                     ((uint32_t)(1<<7))
#define CSI_CSICR1_CLR_STATFIFO                 ((uint32_t)(1<<6))
#define CSI_CSICR1_CLR_RXFIFO                   ((uint32_t)(1<<5))
#define CSI_CSICR1_GCLK_MODE                    ((uint32_t)(1<<4))
#define CSI_CSICR1_INV_DATA                     ((uint32_t)(1<<3))
#define CSI_CSICR1_INV_PCLK                     ((uint32_t)(1<<2))
#define CSI_CSICR1_REDGE                        ((uint32_t)(1<<1))
#define CSI_CSICR1_PIXEL_BIT                    ((uint32_t)(1<<0))

#define CSI_CSICR2_DMA_BURST_TYPE_RFF(n)        ((uint32_t)(((n) & 0x3)<<30))
#define CSI_CSICR2_DMA_BURST_TYPE_SFF(n)        ((uint32_t)(((n) & 0x3)<<28))
#define CSI_CSICR2_DRM                          ((uint32_t)(1<<26))
#define CSI_CSICR2_AFS(n)                       ((uint32_t)(((n) & 0x3)<<24))
#define CSI_CSICR2_SCE                          ((uint32_t)(1<<23))
#define CSI_CSICR2_BTS(n)                       ((uint32_t)(((n) & 0x3)<<19))
#define CSI_CSICR2_LVRM(n)                      ((uint32_t)(((n) & 0x7)<<16))
#define CSI_CSICR2_VSC(n)                       ((uint32_t)(((n) & 0xff)<<8))
#define CSI_CSICR2_HSC(n)                       ((uint32_t)(((n) & 0xff)<<0))
    
#define CSI_CSICR3_FRMCNT(n)                    ((uint32_t)(((n) & 0xffff)<<16))
#define CSI_CSICR3_FRMCNT_RST                   ((uint32_t)(1<<15))
#define CSI_CSICR3_DMA_REFLASH_RFF              ((uint32_t)(1<<14))
#define CSI_CSICR3_DMA_REFLASH_SFF              ((uint32_t)(1<<13))
#define CSI_CSICR3_DMA_REQ_EN_RFF               ((uint32_t)(1<<12))
#define CSI_CSICR3_DMA_REQ_EN_SFF               ((uint32_t)(1<<11))
#define CSI_CSICR3_STATFF_LEVEL(n)              ((uint32_t)(((n) & 0x7)<<8))
#define CSI_CSICR3_HRESP_ERR_EN                 ((uint32_t)(1<<7))
#define CSI_CSICR3_RxFF_LEVEL(n)                ((uint32_t)(((n) & 0x7)<<4))
#define CSI_CSICR3_TWO_8BIT_SENSOR              ((uint32_t)(1<<3))
#define CSI_CSICR3_ZERO_PACK_EN                 ((uint32_t)(1<<2))
#define CSI_CSICR3_ECC_INT_EN                   ((uint32_t)(1<<1))
#define CSI_CSICR3_ECC_AUTO_EN                  ((uint32_t)(1<<0))

#define CSI_CSISR_BASEADDR_CHHANGE_ERROR        ((uint32_t)(1<<28))
#define CSI_CSISR_DMA_FIELD0_DONE               ((uint32_t)(1<<27))
#define CSI_CSISR_DMA_FIELD1_DONE               ((uint32_t)(1<<26))
#define CSI_CSISR_SF_OR_INT                     ((uint32_t)(1<<25))
#define CSI_CSISR_RF_OR_INT                     ((uint32_t)(1<<24))
#define CSI_CSISR_DMA_TSF_DONE_SFF              ((uint32_t)(1<<22))
#define CSI_CSISR_STATFF_INT                    ((uint32_t)(1<<21))
#define CSI_CSISR_DMA_TSF_DONE_FB2              ((uint32_t)(1<<20))
#define CSI_CSISR_DMA_TSF_DONE_FB1              ((uint32_t)(1<<19))
#define CSI_CSISR_RxFF_INT                      ((uint32_t)(1<<18))
#define CSI_CSISR_EOF_INT                       ((uint32_t)(1<<17))
#define CSI_CSISR_SOF_INT                       ((uint32_t)(1<<16))
#define CSI_CSISR_F2_INT                        ((uint32_t)(1<<15))
#define CSI_CSISR_F1_INT                        ((uint32_t)(1<<14))
#define CSI_CSISR_COF_INT                       ((uint32_t)(1<<13))
#define CSI_CSISR_HRESP_ERR_INT                 ((uint32_t)(1<<7))
#define CSI_CSISR_ECC_INT                       ((uint32_t)(1<<1))
#define CSI_CSISR_DRDY                          ((uint32_t)(1<<0))

#define CSI_CSIFBUF_PARA_DEINTERLACE_STRIDE(n)  ((uint32_t)(((n) & 0xffff)<<16))
#define CSI_CSIFBUF_PARA_FBUF_STRIDE(n)         ((uint32_t)(((n) & 0xffff)<<0))

#define CSI_CSIIMAG_PARA_IMAGE_WIDTH(n)         ((uint32_t)(((n) & 0xffff)<<16))
#define CSI_CSIIMAG_PARA_IMAGE_HEIGHT(n)        ((uint32_t)(((n) & 0xffff)<<0))

#define CSI_CSICR18_CSI_ENABLE                  ((uint32_t)(1<<31))
#define CSI_CSICR18_MASK_OPTION(n)              ((uint32_t)(((n) & 0x3)<<18))
#define CSI_CSICR18_AHB_HPROT(n)                ((uint32_t)(((n) & 0xf)<<12))

#define CSI_CSICR18_RGB888A_FORMAT_SEL          ((uint32_t)(1<<10))
#define CSI_CSICR18_BASEADDR_CHANGE_ERROR_IE    ((uint32_t)(1<<9))
#define CSI_CSICR18_LAST_DMA_REQ_SEL            ((uint32_t)(1<<8))
#define CSI_CSICR18_DMA_FIELD1_DONE_IE          ((uint32_t)(1<<7))
#define CSI_CSICR18_FIELD0_DONE_IE              ((uint32_t)(1<<6))
#define CSI_CSICR18_BASEADDR_SWITCH_SEL         ((uint32_t)(1<<5))
#define CSI_CSICR18_BASEADDR_SWITCH_EN          ((uint32_t)(1<<4))
#define CSI_CSICR18_PARALLEL24_EN               ((uint32_t)(1<<3))
#define CSI_CSICR18_DEINTERLACE_EN              ((uint32_t)(1<<2))

#define CSI_CSICR19_DMA_RFIFO_HIGHEST_FIFO_LEVEL(n) ((uint32_t)(((n) & 0xf)<<12))
#endif



#endif //_CSI_41_H_
