//  OV7670 driver using new pinouts for wired breadboard and CSI
//  This version assumes that CSI_41.ino is in the sketch folder
//
#include "CSI_41.h"
#include "OV7670_RegDefs.h"

// Default clock, either 12 or 24
#define DEFAULTCLOCK 12

/*
  const int PinCamVsync = 17; //
  const int PinCamHref  = 16; //
  const int PinCamPpclk = 40; //
  const int PinCamXclk  = 41; //
  const int PinCamD0    = 27; // AD_B1_15
  const int PinCamD1    = 26; //
  const int PinCamD2    = 39; //
  const int PinCamD3    = 38; //
  const int PinCamD4    = 21; //
  const int PinCamD5    = 20; //
  const int PinCamD6    = 23; //
  const int PinCamD7    = 22; // AD_B1_8
*/
// define I2C address for OV7670
static int const camAddress = 0x21; // i2c address of camera's sccb device

struct regval_entry {
  uint8_t reg_num;
  uint8_t value;
};


// Read a single uint8_t from address and return it as a uint8_t
uint8_t cameraReadRegister(uint8_t reg) {
  uint8_t status; 
  Wire.beginTransmission(camAddress);
  Wire.write(reg);
  status = Wire.endTransmission();
  if (status != 0) Serial.printf("Read Register(Write): %x error: %x\n", reg, status); 
  Wire.requestFrom(camAddress, 1);
  uint8_t data = Wire.read();
  status = Wire.endTransmission();
  if (status != 0) Serial.printf("Read Register(request): %x data:%x error: %d\n", reg, data, status); 
  return data;
}

// Writes a single uint8_t (data) into address
#define OV760_DEBUG
#ifdef OV760_DEBUG
typedef struct {
  const char *reg_name;
  unsigned char reg;
} OVREG_TO_NAME_t;

static const OVREG_TO_NAME_t ov_reg_name_table[] =
{
  {"GAIN", 0x00},               //< AGC gain bits 7:0 (9:8 in VREF)
  {"BLUE", 0x01},               //< AWB blue channel gain
  {"RED", 0x02},                //< AWB red channel gain
  {"VREF", 0x03},               //< Vert frame control bits
  {"COM1", 0x04},               //< Common control 1
  {"BAVE", 0x05},               //< U/B average level
  {"GbAVE", 0x06},              //< Y/Gb average level
  {"AECHH", 0x07},              //< Exposure value - AEC 15:10 bits
  {"RAVE", 0x08},               //< V/R average level
  {"COM2", 0x09},               //< Common control 2
  {"PID", 0x0A},                //< Product ID MSB (read-only)
  {"VER", 0x0B},                //< Product ID LSB (read-only)
  {"COM3", 0x0C},               //< Common control 3
  {"COM4", 0x0D},               //< Common control 4
  {"COM5", 0x0E},               //< Common control 5
  {"COM6", 0x0F},               //< Common control 6
  {"AECH", 0x10},               //< Exposure value 9:2
  {"CLKRC", 0x11},              //< Internal clock
  {"COM7", 0x12},               //< Common control 7
  {"COM8", 0x13},               //< Common control 8
  {"COM9", 0x14},               //< Common control 9 - max AGC value
  {"COM10", 0x15},              //< Common control 10
  {"*RSVD*", 0x16},
  {"HSTART", 0x17},             //< Horiz frame start high bits
  {"HSTOP", 0x18},              //< Horiz frame end high bits
  {"VSTART", 0x19},             //< Vert frame start high bits
  {"VSTOP", 0x1A},              //< Vert frame end high bits
  {"PSHFT", 0x1B},              //< Pixel delay select
  {"MIDH", 0x1C},               //< Manufacturer ID high byte
  {"MIDL", 0x1D},               //< Manufacturer ID low byte
  {"MVFP", 0x1E},               //< Mirror / vert-flip enable
  {"LAEC", 0x1F},               //< Reserved
  {"ADCCTR0", 0x20},            //< ADC control
  {"ADCCTR1", 0x21},            //< Reserved
  {"ADCCTR2", 0x22},            //< Reserved
  {"ADCCTR3", 0x23},            //< Reserved
  {"AEW", 0x24},                //< AGC/AEC upper limit
  {"AEB", 0x25},                //< AGC/AEC lower limit
  {"VPT", 0x26},                //< AGC/AEC fast mode op region
  {"BBIAS", 0x27},              //< B channel signal output bias
  {"GbBIAS", 0x28},             //< Gb channel signal output bias
  {"*RSVD*", 0x29},
  {"EXHCH", 0x2A},              //< Dummy pixel insert MSB
  {"EXHCL", 0x2B},              //< Dummy pixel insert LSB
  {"RBIAS", 0x2C},              //< R channel signal output bias
  {"ADVFL", 0x2D},              //< Insert dummy lines MSB
  {"ADVFH", 0x2E},              //< Insert dummy lines LSB
  {"YAVE", 0x2F},               //< Y/G channel average value
  {"HSYST", 0x30},              //< HSYNC rising edge delay
  {"HSYEN", 0x31},              //< HSYNC falling edge delay
  {"HREF", 0x32},               //< HREF control
  {"CHLF", 0x33},               //< Array current control
  {"ARBLM", 0x34},              //< Array ref control - reserved
  {"*RSVD*", 0x35},
  {"*RSVD*", 0x36},
  {"ADC", 0x37},                //< ADC control - reserved
  {"ACOM", 0x38},               //< ADC & analog common - reserved
  {"OFON", 0x39},               //< ADC offset control - reserved
  {"TSLB", 0x3A},               //< Line buffer test option
  {"COM11", 0x3B},              //< Common control 11
  {"COM12", 0x3C},              //< Common control 12
  {"COM13", 0x3D},              //< Common control 13
  {"COM14", 0x3E},              //< Common control 14
  {"EDGE", 0x3F},               //< Edge enhancement adjustment
  {"COM15", 0x40},              //< Common control 15
  {"COM16", 0x41},              //< Common control 16
  {"COM17", 0x42},              //< Common control 17
  {"AWBC1", 0x43},              //< Reserved
  {"AWBC2", 0x44},              //< Reserved
  {"AWBC3", 0x45},              //< Reserved
  {"AWBC4", 0x46},              //< Reserved
  {"AWBC5", 0x47},              //< Reserved
  {"AWBC6", 0x48},              //< Reserved
  {"*RSVD*", 0x49},
  {"*RSVD*", 0x4A},
  {"REG4B", 0x4B},              //< UV average enable
  {"DNSTH", 0x4C},              //< De-noise strength
  {"DM_POS", 0x4D},             // Dummy row position
  {"*RSVD*", 0x4E},
  {"MTX1", 0x4F},               //< Matrix coefficient 1
  {"MTX2", 0x50},               //< Matrix coefficient 2
  {"MTX3", 0x51},               //< Matrix coefficient 3
  {"MTX4", 0x52},               //< Matrix coefficient 4
  {"MTX5", 0x53},               //< Matrix coefficient 5
  {"MTX6", 0x54},               //< Matrix coefficient 6
  {"BRIGHT", 0x55},             //< Brightness control
  {"CONTRAS", 0x56},            //< Contrast control
  {"CONTRAS_CENTER", 0x57},     //< Contrast center
  {"MTXS", 0x58},               //< Matrix coefficient sign
  {"AWBC7", 0x59},              //< AWB Control
  {"AWBC8", 0x5A},              //< AWB control
  {"AWBC9", 0x5B},              //< AWB control
  {"AWBC10", 0x5C},             //< AWB control
  {"AWBC11", 0x5D},             //< AWB control
  {"AWBC12", 0x5E},             //< AWB control
  {"B_LMT", 0x5F},              //< AWB Gain control
  {"R_LMT", 0x60},              //< AWB Gain control
  {"G_LMT", 0x61},              //< AWB Gain control
  {"LCC1", 0x62},               //< Lens correction option 1
  {"LCC2", 0x63},               //< Lens correction option 2
  {"LCC3", 0x64},               //< Lens correction option 3
  {"LCC4", 0x65},               //< Lens correction option 4
  {"LCC5", 0x66},               //< Lens correction option 5
  {"MANU", 0x67},               //< Manual U value
  {"MANV", 0x68},               //< Manual V value
  {"GFIX", 0x69},               //< Fix gain control
  {"GGAIN", 0x6A},              //< G channel AWB gain
  {"DBLV", 0x6B},               //< PLL & regulator control
  {"AWBCTR3", 0x6C},            //< AWB control 3
  {"AWBCTR2", 0x6D},            //< AWB control 2
  {"AWBCTR1", 0x6E},            //< AWB control 1
  {"AWBCTR0", 0x6F},            //< AWB control 0
  {"SCALING_XSC", 0x70},        //< Test pattern X scaling
  {"SCALING_YSC", 0x71},        //< Test pattern Y scaling
  {"SCALING_DCWCTR", 0x72},     //< DCW control
  {"SCALING_PCLK_DIV", 0x73},   //< DSP scale control clock divide
  {"REG74", 0x74},              //< Digital gain control
  {"REG75", 0x75},              //< Edge control
  {"REG76", 0x76},              //< Pixel correction
  {"REG77", 0x77},              //< Offset de-noise
  {"*RSVD*", 0x78},
  {"*RSVD*", 0x79},
  {"SLOP", 0x7A},               //< Gamma curve highest seg slope
  {"GAMA1", 0x7B},              //< Gamma register base (1 of 15)
  {"GAMA2", 0x7C},              //< Gamma register base (1 of 15)
  {"GAMA3", 0x7D},              //< Gamma register base (1 of 15)
  {"GAMA4", 0x7E},              //< Gamma register base (1 of 15)
  {"GAMA5", 0x7F},              //< Gamma register base (1 of 15)
  {"GAMA6", 0x80},              //< Gamma register base (1 of 15)
  {"GAMA7", 0x81},              //< Gamma register base (1 of 15)
  {"GAMA8", 0x82},              //< Gamma register base (1 of 15)
  {"GAMA9", 0x83},              //< Gamma register base (1 of 15)
  {"GAMA10", 0x84},              //< Gamma register base (1 of 15)
  {"GAMA11", 0x85},              //< Gamma register base (1 of 15)
  {"GAMA12", 0x86},              //< Gamma register base (1 of 15)
  {"GAMA13", 0x87},              //< Gamma register base (1 of 15)
  {"GAMA14", 0x88},              //< Gamma register base (1 of 15)
  {"GAMA15", 0x89},              //< Gamma register base (1 of 15)
  {"*RSVD*", 0x8A},
  {"*RSVD*", 0x8B},
  {"?RGB444", 0x8C},             //< RGB 444 control#if 0
  {"*RSVD*", 0x8D},
  {"*RSVD*", 0x8E},
  {"*RSVD*", 0x8F},
  {"*RSVD*", 0x90},
  {"*RSVD*", 0x91},
  {"DM_LNL", 0x92},             //< Dummy line LSB
  {"DM_LNH", 0x93},             //< Dummy line LSB
  {"LCC6", 0x94},               //< Lens correction option 6
  {"LCC7", 0x95},               //< Lens correction option 7
  {"*RSVD*", 0x96},
  {"*RSVD*", 0x97},
  {"*RSVD*", 0x98},
  {"*RSVD*", 0x99},
  {"*RSVD*", 0x9A},
  {"*RSVD*", 0x9B},
  {"*RSVD*", 0x9C},
  {"BD50ST", 0x9D},
  {"BD60ST", 0x9E},
  {"HAECC1", 0x9F},             //< Histogram-based AEC/AGC ctrl 1
  {"HAECC2", 0xA0},             //< Histogram-based AEC/AGC ctrl 2
  {"DSPC3", 0xA1},
  {"SCALING_PCLK_DELAY", 0xA2}, //< Scaling pixel clock delay
  {"*RSVD*", 0xA3},
  {"NT_CTRL", 0xA4},
  {"AECGMAX", 0xA5},            //< 50 Hz banding step limit
  {"LPH", 0xA6},             //< Histogram-based AEC/AGC ctrl 3
  {"UPL", 0xA7},             //< Histogram-based AEC/AGC ctrl 4
  {"TPL", 0xA8},             //< Histogram-based AEC/AGC ctrl 5
  {"TPH", 0xA9},             //< Histogram-based AEC/AGC ctrl 6
  {"NALG", 0xAA},             //< Histogram-based AEC/AGC ctrl 7
  {"*RSVD*", 0xAB},            //< 60 Hz banding step limit
  {"STR-OPT", 0xAC},
  {"STR_R", 0xAD},
  {"STR_G", 0xAE},
  {"STR_B", 0xAF},
  {"*RSVD*", 0xB0},
  {"ABLC1", 0xB1},              //< ABLC enable
  {"*RSVD*", 0xB2},
  {"THL_ST", 0xB3},             //< ABLC target
  {"*RSVD*", 0xB4},
  {"THL_DLT", 0xB5},
  {"*RSVD*", 0xB6},
  {"*RSVD*", 0xB7},
  {"*RSVD*", 0xB8},
  {"*RSVD*", 0xB9},
  {"*RSVD*", 0xBA},
  {"*RSVD*", 0xBB},
  {"*RSVD*", 0xBC},
  {"*RSVD*", 0xBD},
  {"AD-CHB", 0xBE},
  {"AD-CHR", 0xBF},
  {"AD-CHGb", 0xC0},
  {"AD-CHRr", 0xC1},
  {"*RSVD*", 0xC2},
  {"*RSVD*", 0xC3},
  {"*RSVD*", 0xC4},
  {"*RSVD*", 0xC5},
  {"*RSVD*", 0xC6},
  {"*RSVD*", 0xC7},
  {"*RSVD*", 0xC8},
  {"SATCTR", 0xC9}              //< Saturation control
};

#endif

void cameraWriteRegister(uint8_t reg, uint8_t data) {
#ifdef OV760_DEBUG
  Serial.printf("I2C Write: reg: 0x%02x", reg);

  for (uint16_t i = 0; i < (sizeof(ov_reg_name_table) / sizeof(ov_reg_name_table[0])); i++) {
    if (ov_reg_name_table[i].reg == reg) {
      Serial.printf("(%s)", ov_reg_name_table[i].reg_name);
      break;
    }
  }
  Serial.printf(", value = 0x%02x\n", data);
#endif

  Wire.beginTransmission(camAddress);
  Wire.write(reg); Wire.write(data);
  uint8_t status = Wire.endTransmission();
  if (status != 0) {
    Serial.printf("Write Register:%x = %x failed: %d\n", reg, data, status);
  }
}


// simplified test for end of reglist is  reg_num == 0xFF
// This will have to change if we ever get a camera with
// 255 or more registers!
void cameraWriteRegList(struct regval_entry reglist[]) {
  uint16_t index = 0;
  uint8_t reg_addr, reg_val;
  while (reglist[index].reg_num != 0xFF) {
    reg_addr = reglist[index].reg_num;
    reg_val = reglist[index].value;
    cameraWriteRegister(reg_addr, reg_val);
    delayMicroseconds(100);
    index++;
  }
}


// Reset all OV7670 registers to their hardware default values
void cameraReset() {
  cameraWriteRegister(0x12, 0x80); delay(10);
  cameraWriteRegister(0x12, 0); delay(10);
}

// Read and display all of the OV7670 registers
// more compact display  MJB 10/3/2020
void cameraReadAll(uint8_t rvals[]) {
  for (int i = 0; i <= 0xC9; i ++) {
    rvals[i] = cameraReadRegister(i);
  }
}

void cameraShowAll(uint8_t rvals[]) {
  for (int reg = 0; reg <= 0xC9; reg++) {
    Serial.printf("%02X", reg);
    for (uint16_t i = 0; i < (sizeof(ov_reg_name_table) / sizeof(ov_reg_name_table[0])); i++) {
      if (ov_reg_name_table[i].reg == reg) {
        Serial.printf("(%s)", ov_reg_name_table[i].reg_name);
        break;
      }
    }
    Serial.printf(":%02X\n", rvals[reg]);
  }
}

  /*
     https://github.com/ComputerNerd/ov7670-no-ram-arduino-uno/blob/master/ov7670.c
  */
  struct regval_entry regsBayerRGB [] = {
    {REG_COM7, COM7_BAYER},
    {REG_COM13, 0x08}, /* No gamma, magic rsvd bit */
    {REG_COM16, 0x3d}, /* Edge enhancement, denoise */
    {REG_REG76, 0xe1}, /* Pix correction, magic rsvd */
    {0xff, 0xff},  /* END MARKER */
  };

  struct regval_entry regsQQVGA [] = {
    {REG_COM3, COM3_DCWEN}, // enable downsamp/crop/window

    {REG_COM14, 0x1a},  // divide by 4
    {0x72, 0x22},   // downsample by 4
    {0x73, 0xf2},   // divide by 4
    {REG_HSTART, 0x16},
    {REG_HSTOP, 0x04},
    {REG_HREF, 0xa4},
    {REG_VSTART, 0x02},
    {REG_VSTOP, 0x7a},
    {REG_VREF, 0x0a},
    {0xff, 0xff}, /* END MARKER */
  };

  struct regval_entry regsQVGA [] = {
    {REG_COM3, COM3_DCWEN}, // enable downsamp/crop/window

    {REG_COM14, 0x19},
    {0x72, 0x11},
    {0x73, 0xf1},
    {REG_HSTART, 0x16},
    {REG_HSTOP, 0x04},
    {REG_HREF, 0x24},
    {REG_VSTART, 0x02},
    {REG_VSTOP, 0x7a},
    {REG_VREF, 0x0a},
    {0xff, 0xff},  /* END MARKER */
  };

  struct regval_entry regsYUV422 [] = {
    {REG_COM7, 0x0},  /* Selects YUV mode */
    {REG_RGB444, 0},  /* No RGB444 please */
    {REG_COM1, 0},
    {REG_COM15, COM15_R00FF},
    {REG_COM9, 0x6A}, /* 128x gain ceiling; 0x8 is reserved bit */
    {0x4f, 0x80},   /* "matrix coefficient 1" */
    {0x50, 0x80},   /* "matrix coefficient 2" */
    {0x51, 0},    /* vb */
    {0x52, 0x22},   /* "matrix coefficient 4" */
    {0x53, 0x5e},   /* "matrix coefficient 5" */
    {0x54, 0x80},   /* "matrix coefficient 6" */
    {REG_COM13,/*COM13_GAMMA|*/COM13_UVSAT},
    {0xff, 0xff},
  };

  struct regval_entry vga_ov7670[]   = {
    {REG_COM14, 0x18},  // divide by 1   add3ed by MJB 10/18/20
    {0x72, 0x0},   // downsample by 1
    {0x73, 0xf0},   // divide by 1
    {0x15, 0x22},  // No PCLK during H blanking negative-going  VSync
    {REG_HREF, 0xB6}, // was 24
    {0x17, 0x16},   // HSTART
    {0x18, 0x04},   // HSTOP  was 04
    {0x19, 0x01},   // VSTARTwas 02
    {0x1a, 0x79},   // VSTOP  was 7A
    {REG_VREF, 0x3F}, // VREF was 0A
    {0xff, 0xff},   /* END MARKER */
  };

  struct regval_entry rgb565_ov7670[]   = {
    {REG_COM7, COM7_RGB}, /* Selects RGB mode */
    {REG_RGB444, 0},    /* No RGB444 please */
    {REG_COM1, 0x0},
    {REG_COM15, 0xD0},
    {REG_COM9, 0x6A},  /* 128x gain ceiling; 0x8 is reserved bit */
    {0x4f, 0xb3},    /* "matrix coefficient 1" */
    {0x50, 0xb3},    /* "matrix coefficient 2" */
    {0x51, 0},     /* vb */
    {0x52, 0x3d},    /* "matrix coefficient 4" */
    {0x53, 0xa7},    /* "matrix coefficient 5" */
    {0x54, 0xe4},    /* "matrix coefficient 6" */
    {REG_COM13, /*COM13_GAMMA|*/COM13_UVSAT},

    // added a boost in contrast and brightness MJB 10/27/2020
    {REG_BRIGHT, 0X32},  // default is 0
    {REG_CONTRAS, 0x5C}, // default is 0x40

    {0xff, 0xff}, /* END MARKER */
  };


  struct regval_entry ov7670_default_regs[]   = {//from the linux driver
    {REG_COM7, COM7_RESET},
    {REG_TSLB,  0x04},  /* OV */
    {REG_COM7, 0},  /* VGA */
    /*
       Set the hardware window.  These values from OV don't entirely
       make sense - hstop is less than hstart.  But they work...
    */
    {REG_HSTART, 0x16}, {REG_HSTOP, 0x04},
    {REG_HREF, 0xB6}, {REG_VSTART, 0x01},
    {REG_VSTOP, 0x79},  {REG_VREF, 0x3F},

    {REG_COM3, 0},  {REG_COM14, 0},
    /* Mystery scaling numbers */
    {0x70, 0x3a},   {0x71, 0x35},
    {0x72, 0x11},   {0x73, 0xf0},
    {0xa2,/* 0x02 changed to 1*/1}, {REG_COM10, COM10_VS_NEG},
    /* Gamma curve values */

    {0x7a, 0x20},   {0x7b, 0x10},
    {0x7c, 0x1e},   {0x7d, 0x35},
    {0x7e, 0x5a},   {0x7f, 0x69},
    {0x80, 0x76},   {0x81, 0x80},
    {0x82, 0x88},   {0x83, 0x8f},
    {0x84, 0x96},   {0x85, 0xa3},
    {0x86, 0xaf},   {0x87, 0xc4},
    {0x88, 0xd7},   {0x89, 0xe8},

    /* AGC and AEC parameters.  Note we start by disabling those features,
       then turn them only after tweaking the values. */
    {REG_COM8, COM8_FASTAEC | COM8_AECSTEP},
    {REG_GAIN, 0},  {REG_AECH, 0},
    {REG_COM4, 0x40}, /* magic reserved bit */
    {REG_COM9, 0x18}, /* 4x gain + magic rsvd bit */
    {REG_BD50MAX, 0x05},  {REG_BD60MAX, 0x07},
    {REG_AEW, 0x95},  {REG_AEB, 0x33},
    {REG_VPT, 0xe3},  {REG_HAECC1, 0x78},
    {REG_HAECC2, 0x68}, {0xa1, 0x03}, /* magic */
    {REG_HAECC3, 0xd8}, {REG_HAECC4, 0xd8},
    {REG_HAECC5, 0xf0}, {REG_HAECC6, 0x90},
    {REG_HAECC7, 0x94},
    {REG_COM8, COM8_FASTAEC | COM8_AECSTEP | COM8_AGC | COM8_AEC},
    {0x30, 0}, {0x31, 0}, //disable some delays

    // {0xff, 0xff}, /* END MARKER  for testing */

    /* Almost all of these are magic "reserved" values.  */
    {REG_COM5, 0x61}, {REG_COM6, 0x4b},
    {0x16, 0x02},   {REG_MVFP, 0x07},
    {0x21, 0x02},   {0x22, 0x91},
    {0x29, 0x07},   {0x33, 0x0b},
    {0x35, 0x0b},   {0x37, 0x1d},
    {0x38, 0x71},   {0x39, 0x2a},
    {REG_COM12, 0x78},  {0x4d, 0x40},
    {0x4e, 0x20},   {REG_GFIX, 0},
    /*{0x6b, 0x4a},*/   {0x74, 0x10},
    {0x8d, 0x4f},   {0x8e, 0},
    {0x8f, 0},    {0x90, 0},
    {0x91, 0},    {0x96, 0},
    {0x9a, 0},    {0xb0, 0x84},
    {0xb1, 0x0c},   {0xb2, 0x0e},
    {0xb3, 0x82},   {0xb8, 0x0a},

    /* More reserved magic, some of which tweaks white balance */
    {0x43, 0x0a},   {0x44, 0xf0},
    {0x45, 0x34},   {0x46, 0x58},
    {0x47, 0x28},   {0x48, 0x3a},
    {0x59, 0x88},   {0x5a, 0x88},
    {0x5b, 0x44},   {0x5c, 0x67},
    {0x5d, 0x49},   {0x5e, 0x0e},
    {0x6c, 0x0a},   {0x6d, 0x55},
    {0x6e, 0x11},   {0x6f, 0x9e}, /* it was 0x9F "9e for advance AWB" */
    {0x6a, 0x40},   {REG_BLUE, 0x40},
    {REG_RED, 0x60},
    {REG_COM8, COM8_FASTAEC | COM8_AECSTEP | COM8_AGC | COM8_AEC | COM8_AWB},

    /* Matrix coefficients */
    {0x4f, 0x80},   {0x50, 0x80},
    {0x51, 0},    {0x52, 0x22},
    {0x53, 0x5e},   {0x54, 0x80},
    {0x58, 0x9e},

    {REG_COM16, COM16_AWBGAIN}, {REG_EDGE, 0},
    {0x75, 0x05},   {REG_REG76, 0xe1},
    {0x4c, 0},    {0x77, 0x01},
    {REG_COM13, /*0xc3*/0x48},  {0x4b, 0x09},
    {0xc9, 0x60},   /*{REG_COM16, 0x38},*/
    {0x56, 0x40},

    {0x34, 0x11},   {REG_COM11, COM11_EXP | COM11_HZAUTO},
    {0xa4, 0x82/*Was 0x88*/},   {0x96, 0},
    {0x97, 0x30},   {0x98, 0x20},
    {0x99, 0x30},   {0x9a, 0x84},
    {0x9b, 0x29},   {0x9c, 0x03},
    {0x9d, 0x4c},   {0x9e, 0x3f},
    {0x78, 0x04},

    /* Extra-weird stuff.  Some sort of multiplexor register */
    {0x79, 0x01},   {0xc8, 0xf0},
    {0x79, 0x0f},   {0xc8, 0x00},
    {0x79, 0x10},   {0xc8, 0x7e},
    {0x79, 0x0a},   {0xc8, 0x80},
    {0x79, 0x0b},   {0xc8, 0x01},
    {0x79, 0x0c},   {0xc8, 0x0f},
    {0x79, 0x0d},   {0xc8, 0x20},
    {0x79, 0x09},   {0xc8, 0x80},
    {0x79, 0x02},   {0xc8, 0xc0},
    {0x79, 0x03},   {0xc8, 0x40},
    {0x79, 0x05},   {0xc8, 0x30},
    {0x79, 0x26},

    {0xff, 0xff}, /* END MARKER */
  };

  // Start up the camera.  Assume that MCLK is already present
  void cameraBegin(uint16_t camwidth, uint16_t camheight, uint8_t *fb1, uint8_t *fb2) {
    CSI_InitIOMux(); // set up connections to IO Pins
    CSI_InitMCLK(DEFAULTCLOCK);
    cameraClock(DEFAULTCLOCK);  // default clock
    delay(150);
    cameraWriteRegList(ov7670_default_regs);
    delay(10);
    // cameraWriteRegList(vga_ov7670);
    delay(10);
    cameraWriteRegList(rgb565_ov7670);

    CSI_Stop();
    CSI_Setup(camwidth, camheight, fb1, fb2);
    CSI_Start();
  }



  void cameraClock(int mhz) // Gets XCLK from masterfreq in CSI_41
  { // Clock = external-clock * pll_mull(0,4,6 or 8 on bits 6 and 7) / (2*divider_5_lsb+1)
    float r, ratio, bestr;
    uint8_t reg;
    uint16_t bestpll, bestdiv, p, d;
    float const pll[3] = { 4.0f, 6.0f, 8.0f };
    ratio = (float) mhz / (float)masterfreq; // this is what we want to get to... find closest value...

    bestpll = 0, bestdiv = 0; bestr = 1000.0;
    for (p = 0; p < 3; p++) {
      for (d = 0; d < 32; d++) {
        r = pll[p] / (2 * (d + 1));
        if (fabs(r - ratio) < fabs(ratio - bestr)) {
          bestr = r;
          bestpll = p + 1;
          bestdiv = d;
        }
      }
    }
    // CLKRC register: Prescaler divide by 31
    reg = cameraReadRegister(0x11); cameraWriteRegister(0x11, (reg & 0b1000000) | bestdiv);
    // DBLV register: PLL = 6
    reg = cameraReadRegister(0x6B); cameraWriteRegister(0x6B, (reg & 0b00111111) | (bestpll << 6));
    Serial.printf( "cam clock %6.2f with pll=%d and div=%d\n", masterfreq * bestr, bestpll, bestdiv);

  }

  // return true if a new frame is available in the cameraImage buffer.
  bool cameraReady() {
    static  uint32_t lastfc;
    bool r;
    r = (framecount != lastfc);
    lastfc = framecount;
    return r;
  }


  bool cameraSetResFormat(uint8_t res, uint8_t format) {
    uint16_t camwidth, camheight;
    if (cameraResetSettings()) {
#if DEFAULTCLOCK > 12
    cameraClock(DEFAULTCLOCK);  // default clock
#endif
      cameraWriteRegList(ov7670_default_regs);
      delay(10);
      switch (format) {
        default:
        case RGB565:
          cameraWriteRegList(rgb565_ov7670);
          break;
        case BAYERRGB:
          cameraWriteRegList(regsBayerRGB);
          break;
        case YUV422:
          cameraWriteRegList(regsYUV422);
          break;
      }

      switch (res) {
        case VGA_640x480:
          cameraWriteRegList(vga_ov7670);
          camwidth = 640;
          camheight = 480;
          break;
        case QVGA_320x240:
          cameraWriteRegList(regsQVGA);
          camwidth = 320;
          camheight = 240;
          break;
        default:
        case QQVGA_160x120:
          cameraWriteRegList(regsQQVGA);
          camwidth = 160;
          camheight = 120;
          break;
      }

      delay(10);
      CSI_Stop();
      CSI_Setup(camwidth, camheight, fb1, fb2);
      CSI_Start();

      return true;
    } else {
      return false;
    }
  }

  bool cameraResetSettings() {
    cameraWriteRegister(REG_COM7, COM7_RESET);
    delay(500);
    return true;
  }
